/* -*- c++ -*- */
/*
 * Copyright 2021 AOR, LTD.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 *
 * version 210129: First released.
 * version 210203: Add switch: complex output or separated output.
 * version 210408-0: Fix bugs, mistaking recv data.
 */

#define L_BUFSIZ 8192			// change if you cannot execute...

#define TOIQ				// define if you want use gqrx...
#define TOFLOAT

#include <libusb-1.0/libusb.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <signal.h>
#include <cstring>
#include <string>

namespace {
	// Version.
	const std::string vers = "rel210601-0f";

	// Output file name.
#ifdef TOIQ
	const std::string binfile_IQ = "iqdatIQ.bin";
#else
	const std::string binfile_I = "iqdatI.bin";
	const std::string binfile_Q = "iqdatQ.bin";
#endif

	// Conversion fixed point to floating point.
	const float base = 2*2147483648.0f;

	// output streams.
#ifdef TOIQ
	std::ofstream ofs_IQ;
#else
	std::ofstream ofs_I;
	std::ofstream ofs_Q;
#endif

	// USB.
	const int vendor_id = 0x08d0;
	const int product_id = 0xa001;
	const int ep2 = 0x2;
	const int ep6 = 0x86;
	libusb_device_handle * handle;

	// capture start/end command.
	unsigned char st_com[] = {0x5a, 0xa5, 0x00, 0x02, 0x41, 0x53};
	unsigned char en_com[] = {0x5a, 0xa5, 0x00, 0x02, 0x41, 0x45};

	// buffer.
	const int p_atom = 512;
	const int packet_sets = L_BUFSIZ;
	const int packet_len = p_atom*3;
	unsigned char in_buf[packet_sets * packet_len + p_atom];

	// submit backup.
//	bool bkb = false;
	unsigned char bkup[6] = {0, 0, 0, 0, 0, 0};

	// write to disk.
#ifdef TOIQ
	char iqbufs[packet_sets*packet_len];
#else
	char ibufs[packet_sets*packet_len/2];
	char qbufs[packet_sets*packet_len/2];
#endif

	// data definition.
	union {
		int32_t w32;
		short w16[2];
		char b8[4];
	} wI, wQ;

	union {
		float f32;
		int32_t w32;
		char b8[4];
	} fI, fQ;

	// Abandon first packets block.
	bool firstAbandon = false;
	// Detect first available byte.
	bool firstContact = false;

	// Read/Write control.
	int rcnt = -8;
	int wcnt = 0;
	int b_offset = 0;
	bool do_loop = true;

	void sighandler(int signum) {
		std::cerr << "CLOSE." << signum << std::endl;
		do_loop = false;
	}

	int submit_transfer(struct libusb_transfer * transfer_in)
	{
		int r = libusb_submit_transfer(transfer_in);
		if (r != 0) {
			raise(SIGINT);
			std::cerr << "Failure submit, ";
			if (r == LIBUSB_ERROR_NO_DEVICE) {
				std::cerr << "no such device." << std::endl;

			} else if (r == LIBUSB_ERROR_NOT_FOUND) {
				std::cerr << "entity not found." << std::endl;
			} else {
				std::cerr << r << std::endl;
			}
			do_loop = false;
			libusb_cancel_transfer(transfer_in);
			libusb_free_transfer(transfer_in);
			libusb_release_interface(handle, 0);
			libusb_close(handle);
			libusb_exit(nullptr);
			return -1;
		} else {}

		return 0;
	}

	void *save_data(void *)
	{
		rcnt = -8;
		wcnt = 0;
		if (firstAbandon) {
			if (firstContact) {
				while (rcnt < (packet_sets*packet_len)) {
					if (b_offset == 0) {
						if (rcnt < 0) {
							rcnt = 0;
							continue;
						} else if (rcnt < p_atom) {
							wI.b8[3] = in_buf[rcnt+0];
							wI.b8[2] = in_buf[rcnt+1];
							wI.b8[1] = in_buf[rcnt+2];
							wI.b8[0] = in_buf[rcnt+3];
							wQ.b8[3] = in_buf[rcnt+4];
							wQ.b8[2] = in_buf[rcnt+5];
							wQ.b8[1] = in_buf[rcnt+6];
							wQ.b8[0] = in_buf[rcnt+7];
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - p_atom)) {
							wI.b8[3] = in_buf[rcnt+0];
							wI.b8[2] = in_buf[rcnt+1];
							wI.b8[1] = in_buf[rcnt+2];
							wI.b8[0] = in_buf[rcnt+3];
							wQ.b8[3] = in_buf[rcnt+4];
							wQ.b8[2] = in_buf[rcnt+5];
							wQ.b8[1] = in_buf[rcnt+6];
							wQ.b8[0] = in_buf[rcnt+7];
							if ( (in_buf[rcnt+1] & 0x01) == 0x00) {
								rcnt += p_atom;
								continue;
							} else {}
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - 8)) {
							wI.b8[3] = in_buf[rcnt+0];
							wI.b8[2] = in_buf[rcnt+1];
							wI.b8[1] = in_buf[rcnt+2];
							wI.b8[0] = in_buf[rcnt+3];
							wQ.b8[3] = in_buf[rcnt+4];
							wQ.b8[2] = in_buf[rcnt+5];
							wQ.b8[1] = in_buf[rcnt+6];
							wQ.b8[0] = in_buf[rcnt+7];
							if ( (in_buf[rcnt+1] & 0x01) == 0x00) {
								rcnt += p_atom;
								break;
							} else {}
							rcnt += 8;
						} else {
							wI.b8[3] = in_buf[rcnt+0];
							wI.b8[2] = in_buf[rcnt+1];
							wI.b8[1] = in_buf[rcnt+2];
							wI.b8[0] = in_buf[rcnt+3];
							wQ.b8[3] = in_buf[rcnt+4];
							wQ.b8[2] = in_buf[rcnt+5];
							wQ.b8[1] = in_buf[rcnt+6];
							wQ.b8[0] = in_buf[rcnt+7];
							rcnt += 8; // debug use only.
						}
					} else if (b_offset == 2) {
						if (rcnt < 0) {
							wI.b8[3] = bkup[0];
							wI.b8[2] = bkup[1];
							wI.b8[1] = bkup[2];
							wI.b8[0] = bkup[3];
							wQ.b8[3] = bkup[4];
							wQ.b8[2] = bkup[5];
							wQ.b8[1] = in_buf[0];
							wQ.b8[0] = in_buf[1];
							rcnt += 8;
						} else if (rcnt < p_atom) {
							wI.b8[3] = in_buf[rcnt+2];
							wI.b8[2] = in_buf[rcnt+3];
							wI.b8[1] = in_buf[rcnt+4];
							wI.b8[0] = in_buf[rcnt+5];
							wQ.b8[3] = in_buf[rcnt+6];
							wQ.b8[2] = in_buf[rcnt+7];
							wQ.b8[1] = in_buf[rcnt+8];
							wQ.b8[0] = in_buf[rcnt+9];
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - p_atom)) {
							wI.b8[3] = in_buf[rcnt+2];
							wI.b8[2] = in_buf[rcnt+3];
							wI.b8[1] = in_buf[rcnt+4];
							wI.b8[0] = in_buf[rcnt+5];
							wQ.b8[3] = in_buf[rcnt+6];
							wQ.b8[2] = in_buf[rcnt+7];
							wQ.b8[1] = in_buf[rcnt+8];
							wQ.b8[0] = in_buf[rcnt+9];
							if ( (in_buf[rcnt+11] & 0x01) == 0x00) {
								rcnt += p_atom;
								wQ.b8[1] = in_buf[rcnt+8];
								wQ.b8[0] = in_buf[rcnt+9];
							} else {}
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - 8)) {
							wI.b8[3] = in_buf[rcnt+2];
							wI.b8[2] = in_buf[rcnt+3];
							wI.b8[1] = in_buf[rcnt+4];
							wI.b8[0] = in_buf[rcnt+5];
							wQ.b8[3] = in_buf[rcnt+6];
							wQ.b8[2] = in_buf[rcnt+7];
							wQ.b8[1] = in_buf[rcnt+8];
							wQ.b8[0] = in_buf[rcnt+9];
							rcnt += 8;
						} else {
							bkup[0] = in_buf[rcnt+2];
							bkup[1] = in_buf[rcnt+3];
							bkup[2] = in_buf[rcnt+4];
							bkup[3] = in_buf[rcnt+5];
							bkup[4] = in_buf[rcnt+6];
							bkup[5] = in_buf[rcnt+7];
							rcnt += 6; // debug use only.
							break;
						}
					} else if (b_offset == 4) {
						if (rcnt < 0) {
							wI.b8[3] = bkup[0];
							wI.b8[2] = bkup[1];
							wI.b8[1] = bkup[2];
							wI.b8[0] = bkup[3];
							wQ.b8[3] = in_buf[0];
							wQ.b8[2] = in_buf[1];
							wQ.b8[1] = in_buf[2];
							wQ.b8[0] = in_buf[3];
							rcnt += 8;
						} else if (rcnt < p_atom) {
							wI.b8[3] = in_buf[rcnt+4];
							wI.b8[2] = in_buf[rcnt+5];
							wI.b8[1] = in_buf[rcnt+6];
							wI.b8[0] = in_buf[rcnt+7];
							wQ.b8[3] = in_buf[rcnt+8];
							wQ.b8[2] = in_buf[rcnt+9];
							wQ.b8[1] = in_buf[rcnt+10];
							wQ.b8[0] = in_buf[rcnt+11];
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - p_atom)) {
							wI.b8[3] = in_buf[rcnt+4];
							wI.b8[2] = in_buf[rcnt+5];
							wI.b8[1] = in_buf[rcnt+6];
							wI.b8[0] = in_buf[rcnt+7];
							wQ.b8[3] = in_buf[rcnt+8];
							wQ.b8[2] = in_buf[rcnt+9];
							wQ.b8[1] = in_buf[rcnt+10];
							wQ.b8[0] = in_buf[rcnt+11];
							if ( (in_buf[rcnt+13] & 0x01) == 0x00) {
								rcnt += p_atom;
								wQ.b8[3] = in_buf[rcnt+8];
								wQ.b8[2] = in_buf[rcnt+9];
								wQ.b8[1] = in_buf[rcnt+10];
								wQ.b8[0] = in_buf[rcnt+11];
							} else {}
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - 8)) {
							wI.b8[3] = in_buf[rcnt+4];
							wI.b8[2] = in_buf[rcnt+5];
							wI.b8[1] = in_buf[rcnt+6];
							wI.b8[0] = in_buf[rcnt+7];
							wQ.b8[3] = in_buf[rcnt+8];
							wQ.b8[2] = in_buf[rcnt+9];
							wQ.b8[1] = in_buf[rcnt+10];
							wQ.b8[0] = in_buf[rcnt+11];
							rcnt += 8;
						} else {
							bkup[0] = in_buf[rcnt+4];
							bkup[1] = in_buf[rcnt+5];
							bkup[2] = in_buf[rcnt+6];
							bkup[3] = in_buf[rcnt+7];
							rcnt += 4; // debug use only.
							break;
						}
					} else if (b_offset == 6) {
						if (rcnt < 0) {
							wI.b8[3] = bkup[0];
							wI.b8[2] = bkup[1];
							wI.b8[1] = in_buf[0];
							wI.b8[0] = in_buf[1];
							wQ.b8[3] = in_buf[2];
							wQ.b8[2] = in_buf[3];
							wQ.b8[1] = in_buf[4];
							wQ.b8[0] = in_buf[5];
							rcnt += 8;
						} else if (rcnt < p_atom) {
							wI.b8[3] = in_buf[rcnt+6];
							wI.b8[2] = in_buf[rcnt+7];
							wI.b8[1] = in_buf[rcnt+8];
							wI.b8[0] = in_buf[rcnt+9];
							wQ.b8[3] = in_buf[rcnt+10];
							wQ.b8[2] = in_buf[rcnt+11];
							wQ.b8[1] = in_buf[rcnt+12];
							wQ.b8[0] = in_buf[rcnt+13];
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - p_atom)) {
							wI.b8[3] = in_buf[rcnt+6];
							wI.b8[2] = in_buf[rcnt+7];
							wI.b8[1] = in_buf[rcnt+8];
							wI.b8[0] = in_buf[rcnt+9];
							wQ.b8[3] = in_buf[rcnt+10];
							wQ.b8[2] = in_buf[rcnt+11];
							wQ.b8[1] = in_buf[rcnt+12];
							wQ.b8[0] = in_buf[rcnt+13];
							if ( (in_buf[rcnt+15] & 0x01) == 0x00) {
								rcnt += p_atom;
								wI.b8[1] = in_buf[rcnt+8];
								wI.b8[0] = in_buf[rcnt+9];
								wQ.b8[3] = in_buf[rcnt+10];
								wQ.b8[2] = in_buf[rcnt+11];
								wQ.b8[1] = in_buf[rcnt+12];
								wQ.b8[0] = in_buf[rcnt+13];
							} else {}
							rcnt += 8;
						} else if (rcnt < (packet_sets * packet_len - 8)) {
							wI.b8[3] = in_buf[rcnt+6];
							wI.b8[2] = in_buf[rcnt+7];
							wI.b8[1] = in_buf[rcnt+8];
							wI.b8[0] = in_buf[rcnt+9];
							wQ.b8[3] = in_buf[rcnt+10];
							wQ.b8[2] = in_buf[rcnt+11];
							wQ.b8[1] = in_buf[rcnt+12];
							wQ.b8[0] = in_buf[rcnt+13];
							rcnt += 8;
						} else {
							bkup[0] = in_buf[rcnt+6];
							bkup[1] = in_buf[rcnt+7];
							rcnt += 2; // debug use only.
							break;
						}
					} else {}

					// received data processing.
					if ( (wI.w16[0] & 0x8000) == 0x8000) {
						wI.w16[1] = wI.w16[1] | 0x0001;
					} else {
						wI.w16[1] = wI.w16[1] & 0xfffe;
					}
					wI.w16[0] = wI.w16[0] << 1;

					if ( (wQ.w16[0] & 0x8000) == 0x8000) {
						wQ.w16[1] = wQ.w16[1] | 0x0001;
					} else {
						wQ.w16[1] = wQ.w16[1] & 0xfffe;
					}
					wQ.w16[0] = wQ.w16[0] << 1;

					if ((wI.w32 == 0) && (wQ.w32 == 0)) {
						std::cerr << ".";
					}

#ifdef TOIQ
					fI.f32 = wI.w32 / base;
					fQ.f32 = wQ.w32 / base;

					iqbufs[wcnt+0] = fI.b8[0];
					iqbufs[wcnt+1] = fI.b8[1];
					iqbufs[wcnt+2] = fI.b8[2];
					iqbufs[wcnt+3] = fI.b8[3];
					iqbufs[wcnt+4] = fQ.b8[0];
					iqbufs[wcnt+5] = fQ.b8[1];
					iqbufs[wcnt+6] = fQ.b8[2];
					iqbufs[wcnt+7] = fQ.b8[3];

					wcnt += 8;
#else  // TOIQ
#ifdef TOFLOAT
					// FixP to FloatP.
					fI.f32 = wI.w32 / base;
					fQ.f32 = wQ.w32 / base;
					ibufs[wcnt+0] = fI.b8[0];
					ibufs[wcnt+1] = fI.b8[1];
					ibufs[wcnt+2] = fI.b8[2];
					ibufs[wcnt+3] = fI.b8[3];
					qbufs[wcnt+0] = fQ.b8[0];
					qbufs[wcnt+1] = fQ.b8[1];
					qbufs[wcnt+2] = fQ.b8[2];
					qbufs[wcnt+3] = fQ.b8[3];
#else  // TOFLOAT
					// FixP(not to FloatP.)
					// Stored buffer.
					ibufs[wcnt+0] = wI.b8[0];
					ibufs[wcnt+1] = wI.b8[1];
					ibufs[wcnt+2] = wI.b8[2];
					ibufs[wcnt+3] = wI.b8[3];
					qbufs[wcnt+0] = wQ.b8[0];
					qbufs[wcnt+1] = wQ.b8[1];
					qbufs[wcnt+2] = wQ.b8[2];
					qbufs[wcnt+3] = wQ.b8[3];
#endif	// TOFLOAT
					wcnt += 4;
#endif	// TOIQ
				}
#ifdef TOIQ
				ofs_IQ.write(iqbufs, wcnt);
#else
				ofs_I.write(ibufs, wcnt);
				ofs_Q.write(qbufs, wcnt);
#endif

				//  (For debug) inspection packet size.
//				std::cerr <<
//					"bf=" << b_offset << ","
//					"r=" << rcnt << "," <<
//					"w=" << wcnt << std::endl;

			} else {
				// Find/detect first available byte.
				while (true) {
					// Read 2 bytes.
					if ( (in_buf[rcnt+1] & 0x01) == 0x01) {
						// Find.
						b_offset = rcnt;
						// But abandon...
						firstContact = true;
						std::cerr << "FirstContacted, bf=";
						std::cerr << b_offset << std::endl;
						break;
					} else {
						rcnt += 2;
					}
				}
			}
		} else {
			// WORK AROUND, IQ firmare.
			// Abandon first packets block.
			firstAbandon = true;
			std::cerr << "FirstAbandoned" << std::endl;
		}
//		std::cerr << b_offset << "," << rcnt << "\n";

		return nullptr;
	}

	void callback_in(struct libusb_transfer * transfer_in)
	{
		if (transfer_in->status != LIBUSB_TRANSFER_COMPLETED) {
			std::cerr << "transfer->status " << transfer_in->status << std::endl;
		} else {}

//		save_data(nullptr);	// Enable, if no thread.
		pthread_t threadID;	// Enable, if thread.
		if (pthread_create(&threadID, nullptr, save_data, nullptr) != 0) {
			std::cerr << "Failure thread." << std::endl;
		} else {}
		pthread_detach(pthread_self());

		if (do_loop) {
			submit_transfer(transfer_in);
		} else {
			std::cerr << "Capture done." << std::endl;
		}
	}
}

int main()
{
	std::cout << "Version: " << vers << std::endl;

#ifdef TOIQ
	ofs_IQ.open(binfile_IQ, std::ios::out|std::ios::binary|std::ios::trunc);
#else
	ofs_I.open(binfile_I, std::ios::out|std::ios::binary|std::ios::trunc);
	ofs_Q.open(binfile_Q, std::ios::out|std::ios::binary|std::ios::trunc);
#endif

	wI.w32 = wQ.w32 = 0;
	fI.w32 = fQ.w32 = 0;

	memset(in_buf, 0, (packet_sets * packet_len + p_atom));

	// libusb_init()
	if (libusb_init(nullptr) != 0) {
		std::cerr << "Unable to initialize libusb." << std::endl;
		return -1;
	} else {}

	// libusb_open_device_with_vid_pid()
	handle = nullptr;
	try {
		handle = libusb_open_device_with_vid_pid(
			nullptr,
			vendor_id,
			product_id);
		if (handle == nullptr) {
			throw 0;
		} else {}
	}
	catch (int) {
		std::cerr << "unable to open." << std::endl;
		libusb_exit(nullptr);
		return -1;
	}

	// (For debug) inspection packet size.
/*	libusb_device * dev = libusb_get_device(handle);
	int max_size = libusb_get_max_iso_packet_size(
		dev, 0x86);
	std::cerr << "max_iso_packet_size=" << max_size << std::endl;
	max_size = libusb_get_max_packet_size(dev, 0x86);
	std::cerr << "max_packet_size=" << max_size << std::endl;
*/

	// libusb_kernel_driver_active()
	// If a kernel driver is active, you cannot claim the interface,
	// and libusb will be unable to perform I/O.
	if (libusb_kernel_driver_active(handle, 0) != 0) {
		std::cout << "Kernel driver is active." << std::endl;
		std::cout << "So, try detaching the kernel driver." << std::endl;
		// Detach a kernel driver from an interface.
		// If successful, you will then be able to claim the interface and perform I/O.
		// libusb_detach_kernel_driver()
		if (libusb_detach_kernel_driver(handle, 0) != 0) {
			std::cout << "Unable to detach the kernel driver." << std::endl;
			libusb_close(handle);
			libusb_exit(nullptr);
			return -1;
		} else {}
		std::cout << "kernel driver detach done." << std::endl;
	} else {}

	// libusb_claim_interface()
	if (libusb_claim_interface(handle, 0) != 0) {
		std::cerr << "Unable to claim the interface." << std::endl;
		libusb_close(handle);
		libusb_exit(nullptr);
		return -1;
	} else {}

	// libusb_alloc_transfer()
	struct libusb_transfer * transfer_in = nullptr;
	transfer_in  = libusb_alloc_transfer(packet_sets);

	// libusb_fill_iso_transfer()
	memset(in_buf, 0, sizeof(in_buf));
	libusb_fill_iso_transfer(transfer_in,
				 handle,
				 ep6,
				 in_buf,
				 packet_sets * packet_len,
				 packet_sets,
				 callback_in,
				 nullptr,
				 0); // no user data

	// libusb_set_iso_packet_length()
	libusb_set_iso_packet_lengths(transfer_in,
				      packet_len);

	// (For debug) inspection transfer length.
/*	std::cerr << "_transfer_in values1: "
		  << transfer_in->length << ", "
		  << transfer_in->actual_length << ", "
		  << transfer_in->num_iso_packets << std::endl;
*/

	// signal setting.
	struct sigaction sigact;
	sigact.sa_handler = sighandler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	sigaction(SIGINT, &sigact, NULL);
	sigaction(SIGTERM, &sigact, NULL);
	sigaction(SIGQUIT, &sigact, NULL);

	// Send capture start command.
	//
	std::cerr << "Send capture start command." << std::endl;
	int actual_len_st;
	int res_st = libusb_bulk_transfer(
		handle,
		ep2,
		st_com,
		sizeof(st_com),
		&actual_len_st,
		0);
	if (res_st == 0) {
		std::cerr << actual_len_st << std::endl;
	} else {
		std::cerr << "Failure command." << std::endl;
		return -1;
	}


	// libusb_submit_transfer()
	int r = submit_transfer(transfer_in);
	if (r < 0) {
		exit(-2);
	}

	// infinite loop : libusb_handle_events_complated()
	// Terminate when signal receive, SIGINT SIGTERM SIGQUIT.
	while (do_loop) {
		r =  libusb_handle_events_completed(nullptr, nullptr);
//		r =  libusb_handle_events(nullptr);
		if (r < 0){   // negative values are errors
			break;
		}
	}

//	save_data();

	sleep(1);

	//
	// Send capture stop command.
	int actual_len_en;
	int res_en = libusb_bulk_transfer(
		handle,
		ep2,
		en_com,
		sizeof(en_com),
		&actual_len_en,
		0);
	if (res_en == 0) {
		std::cerr << actual_len_en << std::endl;
	}
	std::cerr << "Sent capture end command." << std::endl;

	sleep(1);

	// libusb_cancel_transfer()
	libusb_cancel_transfer(transfer_in);
	std::cerr << "canceled" << std::endl;

	// libusb_free_transfer()
	libusb_free_transfer(transfer_in);
	std::cerr << "freed" << std::endl;

	// libusb_release_interface()
	libusb_release_interface(handle, 0);
	std::cerr << "released" << std::endl;

	// libusb_close()
	libusb_close(handle);
	std::cerr << "closed" << std::endl;

	// libusb_exit()
	libusb_exit(nullptr);
	std::cerr << "exit" << std::endl;

	// file close.
#ifdef TOIQ
	if (ofs_IQ) {
		ofs_IQ.close();
	}
#else
	if (ofs_Q) {
		ofs_Q.close();
	}
	if (ofs_I) {
		ofs_I.close();
	}
#endif

	return 0;
}
